-- GranValora Extension for MoneyMoney
-- Scrapes metal price and portfolio value via mein.granvalora.de
-- Returns metal prices as securities
--
-- Since GranValora unfortunately does not specify the purchase price, the script must call up all transactions individually in order to calculate it manually. If you can live without calculating the profit, comment out `GetPurchases()` in line 186 to speed up the update time significantly
--
-- Username: E-Mail-Address
-- Password: Password
--
-- MIT License
-- Original work Copyright (c) 2022-2025 Silas Meyer
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
-- The above copyright notice and this permission notice shall be included in all
-- copies or substantial portions of the Software.
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
-- SOFTWARE.

local gvConnection
local depotNumber
local host = 'https://mein.granvalora.de'
local gvBankCode = 'GranValora'
local currency = 'EUR'

local commodityWKNs = {
        Gold      = '9655157',
        Silber    = '9653103',
        Platin    = '9665545',
        Palladium = '9665529',
}

WebBanking {
        version = 1.00,
        country = 'de',
        description = 'Include your GranValora portfolio in MoneyMoney',
        url = host .. '/login',
        services = { 'GranValora' }
}

function SupportsBank(protocol, bankCode)
        return protocol == ProtocolWebBanking and bankCode == gvBankCode
end

function InitializeSession(protocol, bankCode, username, reserved, password)
        gvConnection = Connection()
        local html = HTML(gvConnection:get(host .. '/lang?lang=en&came_from=%2Flogin'))
        html:xpath("//div[@id='loginpanel']//input[@name='login']"):attr('value', username)
        html:xpath("//div[@id='loginpanel']//input[@name='password']"):attr('value', password)

        local dashboard = HTML(gvConnection:request(html:xpath("//div[@id='loginpanel']//button[@type='submit']"):click()))

        if gvConnection:getBaseURL() ~= host .. '/' then
                return LoginFailed
                -- return 'Failed to log in. Please check your user credentials.'
        end

        depotNumber = dashboard:xpath('/html/body/header/ul/li[1]/a/span/b'):text():match("%((.+)%)") or 'GranValora'
end

function ListAccounts(knownAccounts)
        local portfolio = {
                name = 'GranValora Portfolio',
                accountNumber = depotNumber,
                subAccount = 'Portfolio',
                currency = currency,
                portfolio = true,
                type = AccountTypePortfolio
        }

        local transactions = {
                name = 'GranValora Transactions',
                accountNumber = depotNumber,
                subAccount = 'Transactions',
                currency = currency,
                portfolio = false,
                type = AccountTypeSavings
        }

        return { portfolio, transactions }
end

function RefreshAccount(account, since)
        if account.type == AccountTypePortfolio then
                return RefreshPortfolio(account, since)
        elseif account.type == AccountTypeSavings then
                return RefreshTransactions(account, since)
        end
end

function FetchStatements (accounts, knownIdentifiers)
  local statements = {}

  -- Load documents page.
  local html = HTML(gvConnection:get(host .. '/documents'))

  local statementTable = html:xpath('//*[@id="content"]/div[2]/div[2]/div/div[2]/table/tbody/tr')
        print('Found ' .. statementTable:length() .. ' statements')

  -- Traverse list of statements.
        statementTable:each(function (index, tr)
    local tds = tr:children()
    local statement = {}
    statement.creationDate = ParseDate(tds:get(1):attr('data-order'))
    statement.name = tds:get(2):text()
    statement.identifier = tds:get(3):text()
    if not knownIdentifiers[statement.identifier] then
      print('Downloading statement from ' .. MM.localizeDate(statement.creationDate) .. '.')
      statement.pdf, _, _, statement.filename = gvConnection:request(tr:xpath('.//a'):click())
      table.insert(statements, statement)
    end
  end)

  return {statements=statements}
end

--- func describing the parsing of a number
---@param numberString string
function ParseNumber(numberString)
        local pattern = '(%d-),-(%d+)%.(%d+)'
        local thousands, hundreds, decimal = numberString:match(pattern)

        if not hundreds or not decimal then
                return nil
        end

        return tonumber(thousands .. hundreds .. '.' .. decimal)
end

function ParseDate(dateString)
        return os.time {
                year = string.sub(dateString, 1, 4),
                month = string.sub(dateString, 5, 6),
                day = string.sub(dateString, 7, 8),
                hour = string.sub(dateString, 10, 11),
                min = string.sub(dateString, 12, 13),
                sec = string.sub(dateString, 14, 15),
        }
end

function RefreshPortfolio(account, since)
        local purchaseSum = {}
        local assets = {}

        function GetPurchases()
                local html = HTML(gvConnection:get(host .. '/transaction'))
                local tableRows = html:xpath('//*[@id="content"]/div[3]/div/div/div[2]/table/tbody/tr')
                local links = html:xpath('//*[@id="content"]/div[3]/div/div/div[2]/table/tbody/tr/td[2]/a')

                print('Found ' .. tableRows:length() .. ' transactions')

                tableRows:each(function(index, _)
                        local link = links:get(index)
                        local url = link:attr('href')
                        local code = link:text()

                        local html = HTML(gvConnection:get(host .. url))
                        local rows = html:xpath('//*[@id="content"]/div[5]/div/div/div[2]/table/tbody/tr')

                        local action = html:xpath('//*[@id="content"]/div[4]/div/div/div/div[2]/div/div/table/tbody/tr[1]/td[2]'):text():match('^%s*(.-)%s*$') == 'Buy' and 1 or -1
                        print('Found ' .. rows:length() .. ' assets were bought/selled in transaction ' .. code)

                        rows:each(function(_, asset)
                                local children = asset:children()
                                local name = children:get(2):text()
                                local amount = ParseNumber(children:get(6):text())

                                purchaseSum[name] = (purchaseSum[name] or 0) + action * amount
                        end)
                end)
        end

        local html = HTML(gvConnection:get('https://mein.granvalora.de/vault/'))
        local tableRows = html:xpath('(//*[@id="collapseOne"])[3]/div/table/tbody/tr[count(td)=4]')
        print('Found ' .. tableRows:length() .. ' assets')

        -- Comment out the following line to speed up the update time significantly
        GetPurchases()

        print('name, quantity, price, amount, purchasePrice')
        for i = 1, tableRows:length() do
                -- {name, quantity, price, amount}
                local row = tableRows:get(i)
                local children = row:children()

                local name = children:get(1):text()
                local wkn = commodityWKNs[name]
                local isin
                local quantity = ParseNumber(children:get(2):text())
                local price = ParseNumber(children:get(3):text())
                local amount = ParseNumber(children:get(4):text())
                local purchasePrice = (purchaseSum[name] or 0) / quantity

                if wkn ~= nil then
                        isin = 'XC000' .. wkn
                end

                print(name, quantity, price, amount, purchasePrice)

                table.insert(assets, {
                        name = name,
                        wkn = wkn,
                        isin = isin,
                        currency = nil,
                        quantity = quantity,
                        price = price,
                        amount = amount,
                        purchasePrice = purchasePrice,
                })
        end

        return { securities = assets }
end

function RefreshTransactions(account, since)
        local ts = GetTransactions()

        local transactions = {}

        local balance = 0;

        for _, t in pairs(ts) do
                balance = balance + t.amount

                if since <= t.date then
                        local transactionCode, bookingKey
                        if t.purpose == 'Buy' then
                                transactionCode = 423 -- Edelmetall-Ankauf
                                bookingKey = 'SEC' -- Wertpapiere (bei Angabe eines Hauptbetrages)
                        elseif string.sub(t.purpose, 1, 2) == 'EA' then
                                transactionCode = 321 -- Depotgebühren
                                bookingKey = 'CHG' -- Gebühren und andere Auslagen
                        elseif t.purpose == 'Sell' then
                                transactionCode = 424 -- Edelmetall-Verkauf
                                bookingKey = 'SEC' -- Wertpapiere (bei Angabe eines Hauptbetrages)
                        elseif t.purpose == 'Invest' then
                                bookingKey = 'EQA' -- Equivalent amount (Gegenwertverrechnung)
                        end

                        table.insert(transactions, {
                                accountNumber = account.accountNumber,
                                bankCode = gvBankCode,
                                amount = t.amount,
                                currency = currency,
                                bookingDate = t.date,
                                valueDate = t.date,
                                purpose = t.purpose .. '\n' .. t.transactionCode,
                                endToEndReference = t.transactionCode,
                                transactionCode = transactionCode,
                                bookingKey = bookingKey,
                                booked = true
                        })
                end
        end

        return {
                balance = balance,
                transactions = transactions,
        }
end

function GetTransactions()
        local html = HTML(gvConnection:get(host .. '/transaction'))
        local tableRows = html:xpath('//*[@id="content"]/div[3]/div/div/div[2]/table/tbody/tr')

        local transactions = {}

        print('Found ' .. tableRows:length() .. ' transactions')

        --tableRows:each(function(index, transaction)

        for i = 1, tableRows:length() do
                -- {name, quantity, price, amount}
                local row = tableRows:get(i)
                local children = row:children()
                local transactionCode = children:get(2):text()
                local date = ParseDate(children:get(3):attr('data-order'))
                local purpose = children:get(4):text()
                local amount = tonumber(children:get(5):attr('data-order'))


                if purpose == 'Sell' then
                        amount = -amount
                end

                table.insert(transactions, {
                        transactionCode = transactionCode,
                        date = date,
                        amount = amount,
                        purpose = purpose
                })

                if purpose == 'Buy' then
                        table.insert(transactions, {
                                transactionCode = transactionCode,
                                date = date + 1,
                                amount = -amount,
                                purpose = 'Invest'
                        })
                end
        end

        return transactions
end

function EndSession()
        gvConnection:get(host .. '/logout_handler')
end

-- SIGNATURE: MCwCFDGGuxSrfblnL6XqXC77cunFx02KAhRVTyXrfTRVleMv1EKzp84Gqu7abA==
